# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from anaconda_project.requirements_registry.registry import RequirementsRegistry
from anaconda_project.requirements_registry.requirement import EnvVarRequirement, UserConfigOverrides

from anaconda_project.internal.test.tmpfile_utils import tmp_local_state_file


def test_user_config_overrides():
    overrides = UserConfigOverrides()
    assert overrides.env_spec_name is None
    overrides = UserConfigOverrides(env_spec_name='foo')
    assert overrides.env_spec_name == 'foo'


def test_find_by_env_var_unknown():
    registry = RequirementsRegistry()
    found = registry.find_requirement_by_env_var(env_var='FOO', options=None)
    assert found is not None
    assert isinstance(found, EnvVarRequirement)
    assert found.env_var == 'FOO'
    assert "EnvVarRequirement(env_var='FOO')" == repr(found)


def test_find_by_service_type_unknown():
    registry = RequirementsRegistry()
    found = registry.find_requirement_by_service_type(service_type='blah', env_var='FOO', options=dict())
    assert found is None


def test_autoguess_encrypted_option():
    def req(env_var, options=None):
        return EnvVarRequirement(registry=RequirementsRegistry(), env_var=env_var, options=options)

    assert not req(env_var='FOO').encrypted
    assert req(env_var='FOO', options=dict(encrypted=True)).encrypted

    assert req(env_var='FOO_PASSWORD').encrypted
    assert req(env_var='FOO_SECRET').encrypted
    assert req(env_var='FOO_SECRET_KEY').encrypted

    assert not req(env_var='FOO_PASSWORD', options=dict(encrypted=False)).encrypted
    assert not req(env_var='FOO_SECRET', options=dict(encrypted=False)).encrypted
    assert not req(env_var='FOO_SECRET_KEY', options=dict(encrypted=False)).encrypted


def test_empty_variable_treated_as_unset():
    requirement = EnvVarRequirement(registry=RequirementsRegistry(), env_var='FOO')
    status = requirement.check_status(dict(FOO=''), tmp_local_state_file(), 'default', UserConfigOverrides())
    assert not status
    assert "Environment variable FOO is not set." == status.status_description
    assert [] == status.errors


def test_requirement_repr():
    requirement = EnvVarRequirement(registry=RequirementsRegistry(), env_var='FOO')
    assert "EnvVarRequirement(env_var='FOO')" == repr(requirement)


def test_requirement_status_repr():
    requirement = EnvVarRequirement(registry=RequirementsRegistry(), env_var='FOO')
    status = requirement.check_status(dict(FOO=''), tmp_local_state_file(), 'default', UserConfigOverrides())
    assert "RequirementStatus(False,'Environment variable FOO is not set.',EnvVarRequirement(env_var='FOO'))" == repr(
        status)


def test_requirement_parse_default():
    null_default = dict(default=None)
    string_default = dict(default="foo")
    int_default = dict(default=42)
    float_default = dict(default=3.14)

    # invalid defaults
    bool_default = dict(default=True)
    list_default = dict(default=[])

    def type_error(value):
        return "default value for variable FOO must be null, a string, or a number, not {value}.".format(value=value)

    problems = []

    EnvVarRequirement._parse_default(null_default, "FOO", problems)
    assert null_default == dict()
    assert problems == []

    EnvVarRequirement._parse_default(string_default, "FOO", problems)
    assert string_default == dict(default="foo")
    assert problems == []

    EnvVarRequirement._parse_default(int_default, "FOO", problems)
    assert int_default == dict(default=42)
    assert problems == []

    EnvVarRequirement._parse_default(float_default, "FOO", problems)
    assert float_default == dict(default=3.14)
    assert problems == []

    EnvVarRequirement._parse_default(bool_default, "FOO", problems)
    assert problems == [type_error(True)]

    problems = []
    EnvVarRequirement._parse_default(list_default, "FOO", problems)
    assert problems == [type_error([])]


def test_requirement_default_as_string():
    no_default = dict()
    string_default = dict(default="foo")
    int_default = dict(default=42)
    float_default = dict(default=3.14)

    req = EnvVarRequirement(registry=RequirementsRegistry(), env_var='FOO', options=no_default)
    assert req.default_as_string is None

    req = EnvVarRequirement(registry=RequirementsRegistry(), env_var='FOO', options=string_default)
    assert req.default_as_string == "foo"

    req = EnvVarRequirement(registry=RequirementsRegistry(), env_var='FOO', options=int_default)
    assert req.default_as_string == "42"

    req = EnvVarRequirement(registry=RequirementsRegistry(), env_var='FOO', options=float_default)
    assert req.default_as_string == "3.14"



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2020, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

from anaconda_project.internal.cli.main import _parse_args_and_run_subcommand
from anaconda_project.internal.test.tmpfile_utils import with_directory_contents_completing_project_file
from anaconda_project.internal.simple_status import SimpleStatus


def _monkeypatch_dockerize(monkeypatch):
    params = {}

    def mock_dockerize(*args, **kwargs):
        params['args'] = args
        params['kwargs'] = kwargs
        return SimpleStatus(success=True, description="Yay")

    monkeypatch.setattr('anaconda_project.project_ops.dockerize', mock_dockerize)
    return params


def _monkeypatch_dockerize_fail(monkeypatch):
    params = {}

    def mock_dockerize(*args, **kwargs):
        params['args'] = args
        params['kwargs'] = kwargs
        return SimpleStatus(success=False, description="Boo")

    monkeypatch.setattr('anaconda_project.project_ops.dockerize', mock_dockerize)
    return params


def test_dockerize(capsys, monkeypatch):
    params = _monkeypatch_dockerize(monkeypatch)

    def check(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'dockerize'])
        assert code == 0

        out, err = capsys.readouterr()
        assert 'Yay\n' == out
        assert '' == err

        assert params['kwargs']['tag'] is None
        assert params['kwargs']['command'] == 'default'
        assert params['kwargs']['builder_image'] == 'conda/s2i-anaconda-project-ubi8:latest'
        assert params['kwargs']['build_args'] == []

    with_directory_contents_completing_project_file(dict(), check)


def test_dockerize_fail(capsys, monkeypatch):
    params = _monkeypatch_dockerize_fail(monkeypatch)

    def check(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'dockerize'])
        assert code == 1

        out, err = capsys.readouterr()
        assert 'Boo\n' == err
        assert '' == out

        assert params['kwargs']['tag'] is None
        assert params['kwargs']['command'] == 'default'
        assert params['kwargs']['builder_image'] == 'conda/s2i-anaconda-project-ubi8:latest'
        assert params['kwargs']['build_args'] == []

    with_directory_contents_completing_project_file(dict(), check)


def test_dockerize_tag(capsys, monkeypatch):
    params = _monkeypatch_dockerize(monkeypatch)

    def check(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'dockerize', '-t', 'dockme:1'])
        assert code == 0

        out, err = capsys.readouterr()
        assert 'Yay\n' == out
        assert '' == err

        assert params['kwargs']['tag'] == 'dockme:1'

    with_directory_contents_completing_project_file(dict(), check)


def test_dockerize_command(capsys, monkeypatch):
    params = _monkeypatch_dockerize(monkeypatch)

    def check(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'dockerize', '--command', 'other-command'])
        assert code == 0

        out, err = capsys.readouterr()
        assert 'Yay\n' == out
        assert '' == err

        assert params['kwargs']['command'] == 'other-command'

    with_directory_contents_completing_project_file(dict(), check)


def test_dockerize_build_args(capsys, monkeypatch):
    params = _monkeypatch_dockerize(monkeypatch)

    def check(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'dockerize', '--', '-e', 'CMD=other', '--run'])
        assert code == 0

        out, err = capsys.readouterr()
        assert 'Yay\n' == out
        assert '' == err

        assert params['kwargs']['build_args'] == ['-e', 'CMD=other', '--run']

    with_directory_contents_completing_project_file(dict(), check)


def test_dockerize_builder_image(capsys, monkeypatch):
    params = _monkeypatch_dockerize(monkeypatch)

    def check(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'dockerize', '--builder-image', 'custom:latest'])
        assert code == 0

        out, err = capsys.readouterr()
        assert 'Yay\n' == out
        assert '' == err

        assert params['kwargs']['builder_image'] == 'custom:latest'

    with_directory_contents_completing_project_file(dict(), check)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
import hashlib
import os

from anaconda_project.local_state_file import LocalStateFile
from anaconda_project.requirements_registry.registry import RequirementsRegistry
from anaconda_project.requirements_registry.requirement import UserConfigOverrides
from anaconda_project.requirements_registry.requirements.download import DownloadRequirement

from anaconda_project.internal.test.tmpfile_utils import with_directory_contents

ENV_VAR = 'DATAFILE'


def test_filename_not_set():
    def check_not_set(dirname):
        local_state = LocalStateFile.load_for_directory(dirname)
        requirement = DownloadRequirement(registry=RequirementsRegistry(),
                                          env_var=ENV_VAR,
                                          url='http://example.com',
                                          filename=ENV_VAR)
        status = requirement.check_status(dict(PROJECT_DIR=dirname), local_state, 'default', UserConfigOverrides())
        assert not status
        assert "Environment variable {} is not set.".format(ENV_VAR) == status.status_description

    with_directory_contents({}, check_not_set)


def test_download_filename_missing():
    def check_missing_filename(dirname):
        local_state = LocalStateFile.load_for_directory(dirname)
        filename = '/data.zip'
        requirement = DownloadRequirement(registry=RequirementsRegistry(),
                                          env_var=ENV_VAR,
                                          url='http://localhost/data.zip',
                                          filename='data.zip')
        status = requirement.check_status({
            ENV_VAR: filename,
            'PROJECT_DIR': dirname
        }, local_state, 'default', UserConfigOverrides())
        assert not status
        assert 'File not found: {}'.format(filename) == status.status_description

    with_directory_contents({}, check_missing_filename)


def make_file_with_checksum():
    datafile = ("column1,column2,column3\n"
                "value11,value12,value13\n"
                "value21,value22,value23\n"
                "value31,value32,value33")
    checksum = hashlib.md5()
    checksum.update(datafile.encode('utf-8'))
    digest = checksum.hexdigest()
    return datafile, digest


def test_download_checksum():
    datafile, digest = make_file_with_checksum()

    def verify_checksum(dirname):
        local_state = LocalStateFile.load_for_directory(dirname)
        filename = os.path.join(dirname, 'data.zip')
        requirement = DownloadRequirement(registry=RequirementsRegistry(),
                                          env_var=ENV_VAR,
                                          url='http://localhost/data.zip',
                                          filename='data.zip',
                                          hash_algorithm='md5',
                                          hash_value=digest)
        status = requirement.check_status({
            ENV_VAR: filename,
            'PROJECT_DIR': dirname
        }, local_state, 'default', UserConfigOverrides())
        assert 'File downloaded to {}'.format(filename) == status.status_description
        assert status

    with_directory_contents({'data.zip': datafile}, verify_checksum)


def test_download_with_no_checksum():
    datafile, digest = make_file_with_checksum()

    def downloaded_file_valid(dirname):
        local_state = LocalStateFile.load_for_directory(dirname)
        filename = os.path.join(dirname, 'data.zip')
        requirement = DownloadRequirement(registry=RequirementsRegistry(),
                                          env_var=ENV_VAR,
                                          url='http://localhost/data.zip',
                                          filename='data.zip')
        status = requirement.check_status({
            ENV_VAR: filename,
            'PROJECT_DIR': dirname
        }, local_state, 'default', UserConfigOverrides())
        assert status
        assert 'File downloaded to {}'.format(filename) == status.status_description

    with_directory_contents({'data.zip': datafile}, downloaded_file_valid)


def test_use_variable_name_for_filename():
    problems = []
    kwargs = DownloadRequirement._parse(varname='FOO', item='http://example.com/', problems=problems)
    assert [] == problems
    assert kwargs['filename'] == 'FOO'
    assert kwargs['url'] == 'http://example.com/'
    assert not kwargs['unzip']


def test_checksum_is_not_a_string():
    problems = []
    kwargs = DownloadRequirement._parse(varname='FOO', item=dict(url='http://example.com/', md5=[]), problems=problems)
    assert ['Checksum value for FOO should be a string not [].'] == problems
    assert kwargs is None


def test_description_is_not_a_string():
    problems = []
    kwargs = DownloadRequirement._parse(varname='FOO',
                                        item=dict(url='http://example.com/', description=[]),
                                        problems=problems)
    assert ["'description' field for download item FOO is not a string"] == problems
    assert kwargs is None


def test_description_property():
    problems = []
    kwargs = DownloadRequirement._parse(varname='FOO',
                                        item=dict(url='http://example.com/', description="hi"),
                                        problems=problems)
    assert [] == problems
    assert kwargs['description'] == 'hi'
    req = DownloadRequirement(RequirementsRegistry(), **kwargs)
    assert req.title == 'FOO'


def test_download_item_is_a_list_not_a_string_or_dict():
    problems = []
    kwargs = DownloadRequirement._parse(varname='FOO', item=[], problems=problems)
    assert ["Download name FOO should be followed by a URL string or a dictionary describing the download."] == problems
    assert kwargs is None


def test_download_item_is_none_not_a_string_or_dict():
    problems = []
    kwargs = DownloadRequirement._parse(varname='FOO', item=None, problems=problems)
    assert ["Download name FOO should be followed by a URL string or a dictionary describing the download."] == problems
    assert kwargs is None


def test_unzip_is_not_a_bool():
    problems = []
    kwargs = DownloadRequirement._parse(varname='FOO',
                                        item=dict(url='http://example.com/', unzip=[]),
                                        problems=problems)
    assert ["Value of 'unzip' for download item FOO should be a boolean, not []."] == problems
    assert kwargs is None


def test_use_unzip_if_url_ends_in_zip():
    problems = []
    kwargs = DownloadRequirement._parse(varname='FOO', item='http://example.com/bar.zip', problems=problems)
    assert [] == problems
    assert kwargs['filename'] == 'bar'
    assert kwargs['url'] == 'http://example.com/bar.zip'
    assert kwargs['unzip']
    req = DownloadRequirement(RequirementsRegistry(), **kwargs)
    assert req.filename == 'bar'
    assert req.url == 'http://example.com/bar.zip'
    assert req.unzip


def test_allow_manual_override_of_use_unzip_if_url_ends_in_zip():
    problems = []
    kwargs = DownloadRequirement._parse(varname='FOO',
                                        item=dict(url='http://example.com/bar.zip', unzip=False),
                                        problems=problems)
    assert [] == problems
    assert kwargs['filename'] == 'bar.zip'
    assert kwargs['url'] == 'http://example.com/bar.zip'
    assert not kwargs['unzip']

    req = DownloadRequirement(RequirementsRegistry(), **kwargs)
    assert req.filename == 'bar.zip'
    assert req.url == 'http://example.com/bar.zip'
    assert not req.unzip


def test_use_unzip_if_url_ends_in_zip_and_filename_does_not():
    problems = []
    kwargs = DownloadRequirement._parse(varname='FOO',
                                        item=dict(url='http://example.com/bar.zip', filename='something'),
                                        problems=problems)
    assert [] == problems
    assert kwargs['filename'] == 'something'
    assert kwargs['url'] == 'http://example.com/bar.zip'
    assert kwargs['unzip']


def test_no_unzip_if_url_ends_in_zip_and_filename_also_does():
    problems = []
    kwargs = DownloadRequirement._parse(varname='FOO',
                                        item=dict(url='http://example.com/bar.zip', filename='something.zip'),
                                        problems=problems)
    assert [] == problems
    assert kwargs['filename'] == 'something.zip'
    assert kwargs['url'] == 'http://example.com/bar.zip'
    assert not kwargs['unzip']



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import codecs
import json
import os
import platform
import pytest
import time
from pprint import pprint

from anaconda_project.env_spec import EnvSpec
from anaconda_project.conda_manager import (CondaManagerError, CondaLockSet)
from anaconda_project import __version__ as version
from anaconda_project.frontend import NullFrontend

from anaconda_project.internal.default_conda_manager import (DefaultCondaManager, _extract_common)
import anaconda_project.internal.pip_api as pip_api
import anaconda_project.internal.conda_api as conda_api

from anaconda_project.internal.test.tmpfile_utils import with_directory_contents
from anaconda_project.internal.test.test_conda_api import monkeypatch_conda_not_to_use_links

if platform.system() == 'Windows':
    PYTHON_BINARY = "python.exe"
    IPYTHON_BINARY = "Scripts\\ipython.exe"
    FLAKE8_BINARY = "Scripts\\flake8.exe"
    # Use a different package from the test env due to weird CI path/env errors
    PYINSTRUMENT_BINARY = "Scripts\\pyinstrument.exe"
else:
    PYTHON_BINARY = "bin/python"
    IPYTHON_BINARY = "bin/ipython"
    FLAKE8_BINARY = "bin/flake8"
    # Use a different package from the test env due to weird CI path/env errors
    PYINSTRUMENT_BINARY = "bin/pyinstrument"

test_spec = EnvSpec(name='myenv', conda_packages=['ipython', 'python=3.6'], pip_packages=['pyinstrument'], channels=[])


def test_current_platform_unsupported_by_env_spec(monkeypatch):
    lock_set = CondaLockSet(package_specs_by_platform={'all': []}, platforms=conda_api.default_platforms)
    spec = EnvSpec(name='myenv',
                   conda_packages=['ipython'],
                   pip_packages=['flake8'],
                   channels=[],
                   platforms=['commodore-64', 'apple-2'],
                   lock_set=lock_set)

    def do_test(dirname):
        envdir = os.path.join(dirname, spec.name)

        manager = DefaultCondaManager(frontend=NullFrontend())

        deviations = manager.find_environment_deviations(envdir, spec)

        error = "Env spec 'myenv' does not support current platform %s (it supports: apple-2, commodore-64)" % \
                conda_api.current_platform()
        assert error == deviations.summary

        with pytest.raises(CondaManagerError) as excinfo:
            manager.fix_environment_deviations(envdir, spec, deviations=deviations)
        assert str(excinfo.value).startswith("Unable to update environment at ")

    with_directory_contents(dict(), do_test)


def test_current_platform_unsupported_by_lock_set(monkeypatch):
    lock_set = CondaLockSet(package_specs_by_platform={'all': []}, platforms=[])
    spec = EnvSpec(name='myenv',
                   conda_packages=['ipython'],
                   pip_packages=['flake8'],
                   channels=[],
                   platforms=conda_api.default_platforms,
                   lock_set=lock_set)

    def do_test(dirname):
        envdir = os.path.join(dirname, spec.name)

        manager = DefaultCondaManager(frontend=NullFrontend())

        deviations = manager.find_environment_deviations(envdir, spec)

        error = "Env spec 'myenv' does not have the current platform %s in the lock file" % conda_api.current_platform()
        assert error == deviations.summary

        with pytest.raises(CondaManagerError) as excinfo:
            manager.fix_environment_deviations(envdir, spec, deviations=deviations)
        assert str(excinfo.value).startswith("Unable to update environment at ")

    with_directory_contents(dict(), do_test)


@pytest.mark.slow
def test_conda_create_and_install_and_remove(monkeypatch):
    monkeypatch_conda_not_to_use_links(monkeypatch)

    spec = test_spec
    assert spec.conda_packages == ('ipython', 'python=3.6')
    assert spec.pip_packages == ('pyinstrument', )

    spec_with_phony_pip_package = EnvSpec(name='myenv',
                                          conda_packages=['ipython'],
                                          pip_packages=['pyinstrument', 'nope_not_a_thing'],
                                          channels=[])
    assert spec_with_phony_pip_package.conda_packages == ('ipython', )
    assert spec_with_phony_pip_package.pip_packages == ('pyinstrument', 'nope_not_a_thing')
    assert spec_with_phony_pip_package.pip_package_names_set == set(('pyinstrument', 'nope_not_a_thing'))

    # package url is supposed to be on a nonexistent port, if it
    # causes a problem we need to mock
    spec_with_bad_url_pip_package = EnvSpec(name='myenv',
                                            conda_packages=['ipython'],
                                            pip_packages=['pyinstrument', 'https://127.0.0.1:24729/nope#egg=phony'],
                                            channels=[])
    assert spec_with_bad_url_pip_package.conda_packages == ('ipython', )
    assert spec_with_bad_url_pip_package.pip_packages == ('pyinstrument', 'https://127.0.0.1:24729/nope#egg=phony')
    assert spec_with_bad_url_pip_package.pip_package_names_set == set(('pyinstrument', 'phony'))

    spec_with_old_ipython = EnvSpec(name='myenv',
                                    conda_packages=['ipython=5.4.1'],
                                    pip_packages=['pyinstrument'],
                                    channels=[])
    assert spec_with_old_ipython.conda_packages == ('ipython=5.4.1', )

    spec_with_bokeh = EnvSpec(name='myenv', conda_packages=['bokeh'], pip_packages=['pyinstrument'], channels=[])
    assert spec_with_bokeh.conda_packages == ('bokeh', )

    spec_with_bokeh_and_old_ipython = EnvSpec(name='myenv',
                                              conda_packages=['bokeh', 'ipython=5.4.1'],
                                              pip_packages=['pyinstrument'],
                                              channels=[])
    assert spec_with_bokeh_and_old_ipython.conda_packages == (
        'bokeh',
        'ipython=5.4.1',
    )

    def do_test(dirname):
        from codecs import open as real_open

        envdir = os.path.join(dirname, spec.name)

        manager = DefaultCondaManager(frontend=NullFrontend())

        is_readonly = dict(readonly=False)

        def mock_open(*args, **kwargs):
            if is_readonly['readonly']:
                raise IOError("did not open")
            return real_open(*args, **kwargs)

        monkeypatch.setattr('codecs.open', mock_open)

        assert not os.path.isdir(envdir)
        assert not os.path.exists(os.path.join(envdir, IPYTHON_BINARY))
        assert not os.path.exists(os.path.join(envdir, FLAKE8_BINARY))
        assert not manager._timestamp_file_up_to_date(envdir, spec)

        deviations = manager.find_environment_deviations(envdir, spec)

        assert set(deviations.missing_packages) == {'python', 'ipython'}
        assert deviations.missing_pip_packages == ('pyinstrument', )

        # with create=False, we won't create the env
        with pytest.raises(CondaManagerError) as excinfo:
            manager.fix_environment_deviations(envdir, spec, deviations, create=False)
            assert 'does not exist' in str(excinfo.value)

        assert not os.path.isdir(envdir)

        # now create the env
        manager.fix_environment_deviations(envdir, spec, deviations)

        assert os.path.isdir(envdir)
        assert os.path.isdir(os.path.join(envdir, "conda-meta"))
        assert os.path.exists(os.path.join(envdir, IPYTHON_BINARY))
        assert os.path.exists(os.path.join(envdir, PYINSTRUMENT_BINARY))

        assert manager._timestamp_file_up_to_date(envdir, spec)
        assert not manager._timestamp_file_up_to_date(envdir, spec_with_phony_pip_package)

        # test bad pip package throws error
        deviations = manager.find_environment_deviations(envdir, spec_with_phony_pip_package)

        assert deviations.missing_packages == ()
        assert deviations.wrong_version_packages == ()
        assert deviations.missing_pip_packages == ('nope_not_a_thing', )

        with pytest.raises(CondaManagerError) as excinfo:
            manager.fix_environment_deviations(envdir, spec_with_phony_pip_package, deviations)
        assert 'Failed to install missing pip packages' in str(excinfo.value)
        assert not manager._timestamp_file_up_to_date(envdir, spec_with_phony_pip_package)

        # test bad url package throws error
        deviations = manager.find_environment_deviations(envdir, spec_with_bad_url_pip_package)

        assert deviations.missing_packages == ()
        assert deviations.wrong_version_packages == ()
        assert deviations.missing_pip_packages == ('phony', )

        with pytest.raises(CondaManagerError) as excinfo:
            manager.fix_environment_deviations(envdir, spec_with_bad_url_pip_package, deviations)
        assert 'Failed to install missing pip packages' in str(excinfo.value)
        assert not manager._timestamp_file_up_to_date(envdir, spec_with_bad_url_pip_package)

        # test we notice wrong ipython version AND missing bokeh AND readonly environment
        is_readonly['readonly'] = True
        deviations = manager.find_environment_deviations(envdir, spec_with_bokeh_and_old_ipython)

        assert deviations.missing_packages == ('bokeh', )
        assert deviations.wrong_version_packages == ('ipython', )
        assert deviations.unfixable
        is_readonly['readonly'] = False

        # test we notice only missing bokeh
        deviations = manager.find_environment_deviations(envdir, spec_with_bokeh)

        assert deviations.missing_packages == ('bokeh', )
        assert deviations.wrong_version_packages == ()
        assert not deviations.unfixable

        # test we notice wrong ipython version and can downgrade
        deviations = manager.find_environment_deviations(envdir, spec_with_old_ipython)

        assert deviations.missing_packages == ()
        assert deviations.wrong_version_packages == ('ipython', )
        assert not deviations.unfixable

        manager.fix_environment_deviations(envdir, spec_with_old_ipython, deviations)

        assert manager._timestamp_file_up_to_date(envdir, spec_with_old_ipython)

        deviations = manager.find_environment_deviations(envdir, spec_with_old_ipython)
        assert deviations.missing_packages == ()
        assert deviations.wrong_version_packages == ()

        # update timestamp; this doesn't re-upgrade because `spec` doesn't
        # specify an ipython version
        assert not manager._timestamp_file_up_to_date(envdir, spec)

        deviations = manager.find_environment_deviations(envdir, spec)

        assert deviations.missing_packages == ()
        assert deviations.wrong_version_packages == ()

        # fix_environment_deviations should be a no-op on readonly envs
        # with no deviations, in particular the time stamp file should
        # not be changed and therefore not be up to date
        is_readonly['readonly'] = True
        manager.fix_environment_deviations(envdir, spec, deviations)
        assert not manager._timestamp_file_up_to_date(envdir, spec)

        # when the environment is readwrite, the timestamp file should
        # be updated
        is_readonly['readonly'] = False
        manager.fix_environment_deviations(envdir, spec, deviations)
        assert manager._timestamp_file_up_to_date(envdir, spec)

        deviations = manager.find_environment_deviations(envdir, spec)
        assert deviations.missing_packages == ()
        assert deviations.wrong_version_packages == ()

        # test that we can remove a package
        assert manager._timestamp_file_up_to_date(envdir, spec)
        time.sleep(1)  # removal is fast enough to break our timestamp resolution
        manager.remove_packages(prefix=envdir, packages=['ipython'])
        assert not os.path.exists(os.path.join(envdir, IPYTHON_BINARY))
        assert not manager._timestamp_file_up_to_date(envdir, spec)

        # test for error removing
        with pytest.raises(CondaManagerError) as excinfo:
            manager.remove_packages(prefix=envdir, packages=['ipython'])
        # different versions of conda word this differently
        message = str(excinfo.value)
        valid_strings = ('no packages found to remove', 'Package not found', "named 'ipython' found to remove",
                         'PackagesNotFoundError:', "is missing from the environment")
        assert any(s in message for s in valid_strings)
        assert not manager._timestamp_file_up_to_date(envdir, spec)

        # test failure to exec pip
        def mock_call_pip(*args, **kwargs):
            raise pip_api.PipError("pip fail")

        monkeypatch.setattr('anaconda_project.internal.pip_api._call_pip', mock_call_pip)

        with pytest.raises(CondaManagerError) as excinfo:
            deviations = manager.find_environment_deviations(envdir, spec)
        assert 'pip failed while listing' in str(excinfo.value)

    with_directory_contents(dict(), do_test)


@pytest.mark.slow
def test_timestamp_file_works(monkeypatch):
    monkeypatch_conda_not_to_use_links(monkeypatch)

    spec = test_spec

    def do_test(dirname):
        envdir = os.path.join(dirname, spec.name)

        manager = DefaultCondaManager(frontend=NullFrontend())

        def print_timestamps(when):
            newest_in_prefix = 0
            for d in manager._timestamp_comparison_directories(envdir):
                try:
                    t = os.path.getmtime(d)
                except Exception:
                    t = 0
                if t > newest_in_prefix:
                    newest_in_prefix = t
            timestamp_fname = manager._timestamp_file(envdir, spec)
            try:
                timestamp_file = os.path.getmtime(timestamp_fname)
            except Exception:
                timestamp_file = 0
            print("%s: timestamp file %d prefix %d diff %g" %
                  (when, timestamp_file, newest_in_prefix, newest_in_prefix - timestamp_file))

        print_timestamps("before env creation")

        assert not os.path.isdir(envdir)
        assert not os.path.exists(os.path.join(envdir, IPYTHON_BINARY))
        assert not os.path.exists(os.path.join(envdir, PYINSTRUMENT_BINARY))
        assert not manager._timestamp_file_up_to_date(envdir, spec)

        deviations = manager.find_environment_deviations(envdir, spec)

        assert set(deviations.missing_packages) == {'python', 'ipython'}
        assert deviations.missing_pip_packages == ('pyinstrument', )
        assert not deviations.ok

        manager.fix_environment_deviations(envdir, spec, deviations)

        print_timestamps("after fixing deviations")

        assert os.path.isdir(envdir)
        assert os.path.isdir(os.path.join(envdir, "conda-meta"))
        assert os.path.exists(os.path.join(envdir, IPYTHON_BINARY))
        assert os.path.exists(os.path.join(envdir, PYINSTRUMENT_BINARY))

        assert manager._timestamp_file_up_to_date(envdir, spec)

        called = []
        from anaconda_project.internal.pip_api import installed as real_pip_installed
        from anaconda_project.internal.conda_api import installed as real_conda_installed

        def traced_pip_installed(*args, **kwargs):
            called.append(("pip_api.installed", args, kwargs))
            return real_pip_installed(*args, **kwargs)

        monkeypatch.setattr('anaconda_project.internal.pip_api.installed', traced_pip_installed)

        def trace_conda_installed(*args, **kwargs):
            called.append(("conda_api.installed", args, kwargs))
            return real_conda_installed(*args, **kwargs)

        monkeypatch.setattr('anaconda_project.internal.conda_api.installed', trace_conda_installed)

        deviations = manager.find_environment_deviations(envdir, spec)

        assert [] == called

        assert deviations.missing_packages == ()
        assert deviations.missing_pip_packages == ()
        assert deviations.ok

        assert manager._timestamp_file_up_to_date(envdir, spec)

        # now modify conda-meta and check that we DO call the package managers
        time.sleep(1.1)  # be sure we are in a new second
        conda_meta_dir = os.path.join(envdir, "conda-meta")
        print("conda-meta original timestamp: %d" % os.path.getmtime(conda_meta_dir))
        inside_conda_meta = os.path.join(conda_meta_dir, "thing.txt")
        with codecs.open(inside_conda_meta, 'w', encoding='utf-8') as f:
            f.write(u"This file should change the mtime on conda-meta\n")
        print("file inside conda-meta %d and conda-meta itself %d" %
              (os.path.getmtime(inside_conda_meta), os.path.getmtime(conda_meta_dir)))
        os.remove(inside_conda_meta)

        print_timestamps("after touching conda-meta")

        assert not manager._timestamp_file_up_to_date(envdir, spec)

        deviations = manager.find_environment_deviations(envdir, spec)

        assert len(called) == 2

        assert deviations.missing_packages == ()
        assert deviations.missing_pip_packages == ()
        # deviations should not be ok (due to timestamp)
        assert not deviations.ok

        assert not manager._timestamp_file_up_to_date(envdir, spec)

        # we want to be sure we update the timestamp file even though
        # there wasn't any actual work to do
        manager.fix_environment_deviations(envdir, spec, deviations)

        print_timestamps("after fixing deviations 2")

        assert manager._timestamp_file_up_to_date(envdir, spec)

    with_directory_contents(dict(), do_test)


def test_timestamp_file_ignores_failed_write(monkeypatch):
    monkeypatch_conda_not_to_use_links(monkeypatch)

    spec = test_spec

    def do_test(dirname):
        from codecs import open as real_open

        envdir = os.path.join(dirname, spec.name)

        manager = DefaultCondaManager(frontend=NullFrontend())

        counts = dict(calls=0)

        def mock_open(*args, **kwargs):
            counts['calls'] += 1
            if counts['calls'] == 1:
                raise IOError("did not open")
            else:
                return real_open(*args, **kwargs)

        monkeypatch.setattr('codecs.open', mock_open)

        # this should NOT throw but also should not write the
        # timestamp file (we ignore errors)
        filename = manager._timestamp_file(envdir, spec)
        assert filename.startswith(envdir)
        assert not os.path.exists(filename)
        manager._write_timestamp_file(envdir, spec)
        assert not os.path.exists(filename)
        # the second time we really wsrite it (this is to prove we
        # are looking at the right filename)
        manager._write_timestamp_file(envdir, spec)
        assert os.path.exists(filename)

        # check on the file contents
        with real_open(filename, 'r', encoding='utf-8') as f:
            content = json.loads(f.read())
            assert dict(anaconda_project_version=version) == content

    with_directory_contents(dict(), do_test)


def test_resolve_dependencies_with_conda_api_mock(monkeypatch):
    def mock_resolve_dependencies(pkgs, platform, channels):
        return [('bokeh', '0.12.4', '0'), ('thing', '1.0', '1')]

    monkeypatch.setattr('anaconda_project.internal.conda_api.resolve_dependencies', mock_resolve_dependencies)

    manager = DefaultCondaManager(frontend=NullFrontend())

    lock_set = manager.resolve_dependencies(['bokeh'], channels=(), platforms=(conda_api.current_platform(), ))
    assert lock_set.package_specs_for_current_platform == ('bokeh=0.12.4=0', 'thing=1.0=1')


@pytest.mark.slow
def test_resolve_dependencies_with_actual_conda():
    manager = DefaultCondaManager(frontend=NullFrontend())

    lock_set = manager.resolve_dependencies(['bokeh'], channels=(), platforms=(conda_api.current_platform(), ))
    specs = lock_set.package_specs_for_current_platform
    pprint(specs)
    names = [conda_api.parse_spec(spec).name for spec in specs]
    assert 'bokeh' in names
    assert len(specs) > 5  # 5 is an arbitrary number of deps that surely bokeh has


def test_resolve_dependencies_with_conda_api_mock_raises_error(monkeypatch):
    def mock_resolve_dependencies(pkgs, platform, channels):
        raise conda_api.CondaError("nope")

    monkeypatch.setattr('anaconda_project.internal.conda_api.resolve_dependencies', mock_resolve_dependencies)

    manager = DefaultCondaManager(frontend=NullFrontend())

    with pytest.raises(CondaManagerError) as excinfo:
        manager.resolve_dependencies(['bokeh'], channels=(), platforms=(conda_api.current_platform(), ))

    assert 'Error resolving for' in str(excinfo.value)


def test_installed_version_comparison(monkeypatch):
    def check(dirname):
        prefix = os.path.join(dirname, "myenv")
        os.makedirs(os.path.join(prefix, 'conda-meta'))

        def mock_installed(prefix):
            return {'bokeh': ('bokeh', '0.12.4', '1')}

        monkeypatch.setattr('anaconda_project.internal.conda_api.installed', mock_installed)

        spec_with_matching_bokeh = EnvSpec(name='myenv',
                                           conda_packages=['bokeh=0.12.4=1'],
                                           pip_packages=[],
                                           channels=[])
        spec_with_more_vague_bokeh = EnvSpec(name='myenv', conda_packages=['bokeh=0.12'], pip_packages=[], channels=[])
        spec_with_unspecified_bokeh = EnvSpec(name='myenv', conda_packages=['bokeh'], pip_packages=[], channels=[])
        spec_with_wrong_version_bokeh = EnvSpec(name='myenv',
                                                conda_packages=['bokeh=0.12.3'],
                                                pip_packages=[],
                                                channels=[])
        spec_with_wrong_build_bokeh = EnvSpec(name='myenv',
                                              conda_packages=['bokeh=0.12.4=0'],
                                              pip_packages=[],
                                              channels=[])

        manager = DefaultCondaManager(frontend=NullFrontend())

        deviations = manager.find_environment_deviations(prefix, spec_with_matching_bokeh)
        assert deviations.missing_packages == ()
        assert deviations.wrong_version_packages == ()

        deviations = manager.find_environment_deviations(prefix, spec_with_more_vague_bokeh)
        assert deviations.missing_packages == ()
        assert deviations.wrong_version_packages == ()

        deviations = manager.find_environment_deviations(prefix, spec_with_unspecified_bokeh)
        assert deviations.missing_packages == ()
        assert deviations.wrong_version_packages == ()

        deviations = manager.find_environment_deviations(prefix, spec_with_wrong_version_bokeh)
        assert deviations.missing_packages == ()
        assert deviations.wrong_version_packages == ('bokeh', )

        deviations = manager.find_environment_deviations(prefix, spec_with_wrong_build_bokeh)
        assert deviations.missing_packages == ()
        assert deviations.wrong_version_packages == ('bokeh', )

    with_directory_contents(dict(), check)


def test_extract_common():
    resolve_results = {
        'linux-32': ['linux-32-only', 'linux-only', 'unix-only', 'common'],
        'linux-64': ['linux-64-only', 'linux-only', 'unix-only', 'common'],
        'win-32': ['win-32-only', 'win-only', 'common'],
        'win-64': ['win-64-only', 'win-only', 'common'],
        'osx-64': ['osx-64-only', 'osx-only', 'unix-only', 'common']
    }
    factored = _extract_common(resolve_results)

    assert {
        'all': ['common'],
        'linux': ['linux-only'],
        'unix': ['unix-only'],
        'win': ['win-only'],
        'osx-64': ['osx-64-only', 'osx-only'],
        'linux-32': ['linux-32-only'],
        'linux-64': ['linux-64-only'],
        'win-32': ['win-32-only'],
        'win-64': ['win-64-only']
    } == factored


def test_extract_common_empty_deps():
    resolve_results = {}
    factored = _extract_common(resolve_results)

    assert {} == factored


def test_extract_common_just_one_platform():
    resolve_results = {'linux-64': ['a', 'b']}
    factored = _extract_common(resolve_results)

    assert {'linux-64': ['a', 'b']} == factored


def test_extract_common_nothing_in_common():
    resolve_results = {'linux-64': ['a', 'b'], 'linux-32': ['c', 'd']}
    factored = _extract_common(resolve_results)

    assert {'linux-64': ['a', 'b'], 'linux-32': ['c', 'd']} == factored


def test_extract_common_only_bits_differ():
    resolve_results = {'linux-64': ['a', 'b'], 'linux-32': ['a', 'b', 'c']}
    factored = _extract_common(resolve_results)

    assert {'linux': ['a', 'b'], 'linux-32': ['c']} == factored


def test_extract_common_only_unix():
    resolve_results = {'linux-64': ['a', 'b'], 'linux-32': ['a', 'b', 'c'], 'osx-64': ['a', 'b']}
    factored = _extract_common(resolve_results)

    assert {'unix': ['a', 'b'], 'linux-32': ['c']} == factored


def test_extract_common_unpopular_unix():
    resolve_results = {'linux-64': ['a', 'b'], 'linux-32': ['a', 'b', 'c'], 'osx-32': ['a', 'b']}
    factored = _extract_common(resolve_results)

    assert {'unix': ['a', 'b'], 'linux-32': ['c']} == factored



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Abstract high-level interface to Conda."""
from __future__ import absolute_import

import codecs
import glob
import os
import shutil
import subprocess

from anaconda_project.conda_manager import (CondaManager, CondaEnvironmentDeviations, CondaLockSet, CondaManagerError)
import anaconda_project.internal.conda_api as conda_api
import anaconda_project.internal.pip_api as pip_api
import anaconda_project.internal.makedirs as makedirs

from anaconda_project import __version__ as version


def _refactor_common_packages(existing_sets, include_predicate, factored_name):
    # For items in existing_sets included by include_predicate,
    # try to factor out common items into factored_name.
    # include_predicate takes keys from existing_sets as param.
    factorable_names = [name for name in existing_sets.keys() if include_predicate(name)]
    unfactorable_names = [name for name in existing_sets.keys() if not include_predicate(name)]

    if len(factorable_names) < 2:
        # need at least two things to have a DRY problem; we want to
        # keep things specific unless there's a reason to refactor.
        return existing_sets

    specs = [existing_sets[name] for name in factorable_names]
    factored = set.intersection(*specs)

    if len(factored) == 0:
        # nothing in common amongst relevant things
        return existing_sets

    result = dict()
    for name in factorable_names:
        remaining = existing_sets[name] - factored
        if len(remaining) > 0:
            result[name] = remaining

    for name in unfactorable_names:
        result[name] = existing_sets[name]

    result[factored_name] = factored

    return result


def _extract_common(by_platform):
    # convert from dict of platform-to-lists to platform-to-sets
    result = {name: set(values) for (name, values) in by_platform.items()}

    # linux-64,linux-32 => linux, etc.
    platform_names = set([conda_api.parse_platform(platform)[0] for platform in by_platform.keys()])
    for name in platform_names:
        result = _refactor_common_packages(result, lambda p: p.startswith("%s-" % name), name)

    # is it a candidate for the "unix" grouping
    def is_unix(platform_or_platform_name):
        for unix_name in conda_api.unix_platform_names:
            if platform_or_platform_name.startswith(unix_name):
                return True
        return False

    # if we have "linux","linux-64","osx-64", we should consider "linux","osx-64"
    # and leave out "linux-64".
    def is_most_general(platform_or_platform_name):
        if '-' in platform_or_platform_name:
            platform_name = conda_api.parse_platform(platform_or_platform_name)[0]
            if platform_name in result:
                return False

        if 'unix' in result and is_unix(platform_or_platform_name) and platform_or_platform_name != 'unix':
            return False

        # we don't call this after adding 'all'
        assert platform_or_platform_name != 'all'

        return True

    # linux*, osx* => unix
    result = _refactor_common_packages(result, lambda p: is_unix(p) and is_most_general(p), "unix")

    # everything => all
    result = _refactor_common_packages(result, lambda p: is_most_general(p), "all")

    return {name: sorted(list(value)) for (name, value) in result.items()}


class DefaultCondaManager(CondaManager):
    def __init__(self, frontend):
        self._frontend = frontend

    def _log_info(self, line):
        if self._frontend is not None:
            self._frontend.info(line)

    # uncomment this if you want to use it, not used for now.
    # def _log_error(self, line):
    #     if self._frontend is not None:
    #         self._frontend.error(line)

    def _on_stdout(self, data):
        if self._frontend is not None:
            self._frontend.partial_info(data)

    def _on_stderr(self, data):
        if self._frontend is not None:
            self._frontend.partial_error(data)

    def _cache_directory(self, prefix):
        return os.path.join(prefix, "var", "cache", "anaconda-project")

    def _test_writable_file(self, prefix):
        return os.path.join(self._cache_directory(prefix), "status")

    def _force_readonly_file(self, prefix, parent=False):
        if parent:
            prefix = os.path.dirname(prefix)
        return os.path.join(prefix, '.readonly')

    def _timestamp_file(self, prefix, spec):
        return os.path.join(self._cache_directory(prefix), "env-specs", spec.locked_hash)

    def _timestamp_comparison_directories(self, prefix):
        # this is a little bit heuristic; we are trying to detect
        # if any packages are installed or removed. This may need
        # to become more comprehensive.  We don't want to check
        # directories that would change at runtime like /var/run,
        # and we need this to be reasonably fast (so we can't do a
        # full directory walk or something). Remember that on
        # Linux at least a new mtime on a directory means
        # _immediate_ child directory entries were added or
        # removed, changing the files themselves or the files in
        # subdirs will not affect mtime. Windows may be a bit
        # different.

        # Linux
        dirs = list(glob.iglob(os.path.join(prefix, "lib", "python*", "site-packages")))
        dirs.append(os.path.join(prefix, "bin"))
        dirs.append(os.path.join(prefix, "lib"))
        # Windows
        dirs.append(os.path.join(prefix, "Lib", "site-packages"))
        dirs.append(os.path.join(prefix, "Library", "bin"))
        dirs.append(os.path.join(prefix, "Scripts"))
        # conda-meta
        dirs.append(os.path.join(prefix, "conda-meta"))

        return dirs

    def _timestamp_file_up_to_date(self, prefix, spec):
        # The goal here is to return False if 1) the env spec
        # has changed (different hash) or 2) the environment has
        # been modified (e.g. by pip or conda).

        filename = self._timestamp_file(prefix, spec)
        try:
            stamp_mtime = os.path.getmtime(filename)
        except OSError:
            return False

        dirs = self._timestamp_comparison_directories(prefix)

        for d in dirs:
            try:
                d_mtime = os.path.getmtime(d)
            except OSError:
                d_mtime = 0
            # When we write the timestamp, we put it 1s in the
            # future, so we want >= here (if the d_mtime has gone
            # into the future from when we wrote the timestamp,
            # the directory has changed).
            if d_mtime >= stamp_mtime:
                return False

        return True

    def _write_a_file(self, filename):
        try:
            makedirs.makedirs_ok_if_exists(os.path.dirname(filename))
            with codecs.open(filename, 'w', encoding='utf-8') as f:
                # we don't read the contents of the file for now, but
                # recording the version in it in case in the future
                # that is useful. We need to write something to the
                # file to bump its mtime if it already exists...
                f.write('{"anaconda_project_version": "%s"}\n' % version)
            return True
        except (IOError, OSError):
            # ignore errors because this is just an optimization, if we
            # fail we will survive
            return False

    def _is_environment_writable(self, prefix):
        if (os.path.exists(self._force_readonly_file(prefix))
                or os.path.exists(self._force_readonly_file(prefix, parent=True))):
            return False
        filename = self._test_writable_file(prefix)
        return self._write_a_file(filename)

    def _write_timestamp_file(self, prefix, spec):
        filename = self._timestamp_file(prefix, spec)
        if self._write_a_file(filename):
            # set the timestamp 1s in the future, which guarantees
            # it doesn't have the same mtime as any files in the
            # environment changed by us; if another process
            # changes some files during the current second, then
            # we would not notice those changes. The alternative
            # is that we falsely believe we changed things
            # ourselves. Ultimately clock resolution keeps us from
            # perfection here without some sort of cross-process
            # locking.
            actual_time = os.path.getmtime(filename)
            next_tick_time = actual_time + 1
            os.utime(filename, (next_tick_time, next_tick_time))

    def resolve_dependencies(self, package_specs, channels, platforms):
        by_platform = {}

        current = conda_api.current_platform()
        resolve_for_platforms = list(platforms)
        # always resolve "current" first because it's confusing if
        # an error says resolution failed on another platform when
        # the real issue is that resolution will fail on all platforms.
        if current in resolve_for_platforms:
            resolve_for_platforms.remove(current)
            resolve_for_platforms = [current] + resolve_for_platforms
        for conda_platform in resolve_for_platforms:
            try:
                self._log_info("Resolving conda packages for %s" % conda_platform)
                deps = conda_api.resolve_dependencies(pkgs=package_specs, platform=conda_platform, channels=channels)
            except conda_api.CondaError as e:
                raise CondaManagerError("Error resolving for {}: {}".format(conda_platform, str(e)))
            locked_specs = ["%s=%s=%s" % dep for dep in deps]
            by_platform[conda_platform] = sorted(locked_specs)

        by_platform = _extract_common(by_platform)

        lock_set = CondaLockSet(package_specs_by_platform=by_platform, platforms=resolve_for_platforms)
        return lock_set

    def _find_conda_deviations(self, prefix, env_spec):
        try:
            installed = conda_api.installed(prefix)
        except conda_api.CondaError as e:
            raise CondaManagerError("Conda failed while listing installed packages in %s: %s" % (prefix, str(e)))

        missing = set()
        wrong_version = set()

        for spec_string in env_spec.conda_packages_for_create:
            spec = conda_api.parse_spec(spec_string)
            name = spec.name

            if name not in installed:
                missing.add(name)
            else:

                def version_match(wanted, installed):
                    if wanted == installed:
                        return True
                    else:
                        return installed.startswith(wanted + ".")

                # The only constraint we are smart enough to understand is
                # the one we put in the lock file, which is plain =.
                # We can't do version comparisons, which is a bug.
                # We won't notice if non-= constraints are unmet.
                (_, installed_version, installed_build) = installed[name]
                if spec.exact_version is not None and not version_match(spec.exact_version, installed_version):
                    wrong_version.add(name)
                elif spec.exact_build_string is not None and not version_match(spec.exact_build_string,
                                                                               installed_build):
                    wrong_version.add(name)

        return (sorted(list(missing)), sorted(list(wrong_version)))

    def _find_pip_missing(self, prefix, spec):
        # this is an important optimization to avoid a slow "pip
        # list" operation if the project has no pip packages
        if len(spec.pip_package_names_set) == 0:
            return []

        try:
            installed = pip_api.installed(prefix)
        except pip_api.PipError as e:
            raise CondaManagerError("pip failed while listing installed packages in %s: %s" % (prefix, str(e)))

        # TODO: we don't verify that the environment contains the right versions
        # https://github.com/Anaconda-Server/anaconda-project/issues/77

        missing = set()

        for name in spec.pip_package_names_set:
            if name not in installed:
                missing.add(name)

        return sorted(list(missing))

    def _broken_lock_set_error(self, spec):
        error = None
        if spec.lock_set is not None and spec.lock_set.enabled:
            # We have to check this first, because getting our package list
            # is not valid if we don't have platform support.
            current_platform = conda_api.current_platform()
            if current_platform not in spec.platforms:
                error = "Env spec '%s' does not support current platform %s (it supports: %s)" % \
                        (spec.name, current_platform, ", ".join(spec.platforms))
            elif not spec.lock_set.supports_current_platform:
                error = "Env spec '%s' does not have the current platform %s in the lock file" % \
                        (spec.name, current_platform)
        return error

    def find_environment_deviations(self, prefix, spec):
        broken_lock_set = self._broken_lock_set_error(spec)
        if broken_lock_set is not None:
            return CondaEnvironmentDeviations(summary=broken_lock_set,
                                              missing_packages=(),
                                              wrong_version_packages=(),
                                              missing_pip_packages=(),
                                              wrong_version_pip_packages=(),
                                              broken=True,
                                              unfixable=True)

        if not os.path.isdir(os.path.join(prefix, 'conda-meta')):
            return CondaEnvironmentDeviations(summary="'%s' doesn't look like it contains a Conda environment yet." %
                                              (prefix),
                                              missing_packages=tuple(spec.conda_package_names_for_create_set),
                                              wrong_version_packages=(),
                                              missing_pip_packages=tuple(spec.pip_package_names_for_create_set),
                                              wrong_version_pip_packages=(),
                                              broken=True)

        broken = unfixable = False
        if self._timestamp_file_up_to_date(prefix, spec):
            conda_missing = []
            conda_wrong_version = []
            pip_missing = []
        else:
            (conda_missing, conda_wrong_version) = self._find_conda_deviations(prefix, spec)
            pip_missing = self._find_pip_missing(prefix, spec)
            broken = self._is_environment_writable(prefix)
            # For readonly environments, do not enforce the writing of the timestamp.
            # But mark other deviations as unfixable
            unfixable = not broken and (conda_missing or conda_wrong_version or pip_missing)

        all_missing_string = ", ".join(conda_missing + pip_missing)
        all_wrong_version_string = ", ".join(conda_wrong_version)

        if all_missing_string != "" and all_wrong_version_string != "":
            summary = "Conda environment is missing packages: %s and has wrong versions of: %s" % (
                all_missing_string, all_wrong_version_string)
        elif all_missing_string != "":
            summary = "Conda environment is missing packages: %s" % all_missing_string
        elif all_wrong_version_string != "":
            summary = "Conda environment has wrong versions of: %s" % all_wrong_version_string
        elif broken:
            summary = "Conda environment needs to be marked as up-to-date"
        else:
            summary = "OK"
        if unfixable:
            summary += " and the environment is read-only"
        return CondaEnvironmentDeviations(summary=summary,
                                          missing_packages=conda_missing,
                                          wrong_version_packages=conda_wrong_version,
                                          missing_pip_packages=pip_missing,
                                          wrong_version_pip_packages=(),
                                          broken=broken,
                                          unfixable=unfixable)

    def fix_environment_deviations(self, prefix, spec, deviations=None, create=True):
        if deviations is None:
            deviations = self.find_environment_deviations(prefix, spec)

        if deviations.unfixable:
            raise CondaManagerError("Unable to update environment at %s" % prefix)

        conda_meta = os.path.join(prefix, 'conda-meta')
        packed = os.path.join(conda_meta, '.packed')
        install_pip = True

        if os.path.isdir(conda_meta) and os.path.exists(packed):
            with open(packed) as f:
                packed_arch = f.read().strip()

            matched = packed_arch == conda_api.current_platform()
            if matched:
                if 'win' in conda_api.current_platform():
                    unpack_script = ['python', os.path.join(prefix, 'Scripts', 'conda-unpack-script.py')]

                else:
                    unpack_script = os.path.join(prefix, 'bin', 'conda-unpack')

                try:
                    subprocess.check_call(unpack_script)
                    os.remove(packed)
                    install_pip = False
                except (subprocess.CalledProcessError, OSError) as e:
                    self._log_info('Warning: conda-unpack could not be run: \n{}\n'
                                   'The environment will be recreated.'.format(str(e)))
                    create = True
                    shutil.rmtree(prefix)

            else:
                self._log_info('Warning: The unpacked env does not match the current architecture. '
                               'It will be recreated.')
                create = True
                shutil.rmtree(prefix)

        if os.path.isdir(conda_meta):
            to_update = list(set(deviations.missing_packages + deviations.wrong_version_packages))
            if len(to_update) > 0:
                specs = spec.specs_for_conda_package_names(to_update)
                assert len(specs) == len(to_update)
                spec.apply_pins(prefix, specs)
                try:
                    conda_api.install(prefix=prefix,
                                      pkgs=specs,
                                      channels=spec.channels,
                                      stdout_callback=self._on_stdout,
                                      stderr_callback=self._on_stderr)
                except conda_api.CondaError as e:
                    raise CondaManagerError("Failed to install packages: {}: {}".format(", ".join(specs), str(e)))
                finally:
                    spec.remove_pins(prefix)
        elif create:
            # Create environment from scratch

            command_line_packages = set(spec.conda_packages_for_create)

            try:
                conda_api.create(prefix=prefix,
                                 pkgs=list(command_line_packages),
                                 channels=spec.channels,
                                 stdout_callback=self._on_stdout,
                                 stderr_callback=self._on_stderr)
            except conda_api.CondaError as e:
                raise CondaManagerError("Failed to create environment at %s: %s" % (prefix, str(e)))
        else:
            raise CondaManagerError("Conda environment at %s does not exist" % (prefix))

        # now add pip if needed
        missing = list(deviations.missing_pip_packages)
        if (len(missing) > 0) and install_pip:
            specs = spec.specs_for_pip_package_names(missing)
            assert len(specs) == len(missing)
            try:
                pip_api.install(prefix=prefix,
                                pkgs=specs,
                                stdout_callback=self._on_stdout,
                                stderr_callback=self._on_stderr)
            except pip_api.PipError as e:
                raise CondaManagerError("Failed to install missing pip packages: {}: {}".format(
                    ", ".join(missing), str(e)))

        # write a file to tell us we can short-circuit next time
        self._write_timestamp_file(prefix, spec)

    def remove_packages(self, prefix, packages, pip=False):
        if pip:
            try:
                pip_api.remove(prefix, packages, stdout_callback=self._on_stdout, stderr_callback=self._on_stderr)
            except pip_api.PipError as e:
                raise CondaManagerError('Failed to remove pip packages from {}: {}'.format(prefix, str(e)))
        else:
            try:
                conda_api.remove(prefix, packages, stdout_callback=self._on_stdout, stderr_callback=self._on_stderr)
            except conda_api.CondaError as e:
                raise CondaManagerError("Failed to remove packages from %s: %s" % (prefix, str(e)))
